import sys
import subprocess

version = '0.4.1'

def syscall(cmd, allow_fail=False, verbose=False):
    if verbose:
        print('syscall:', cmd, flush=True)
    try:
        subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as error:
        errors = error.output.decode()
        print('The following command failed with exit code', error.returncode, file=sys.stderr)
        print(cmd, file=sys.stderr)
        print('\nThe output was:\n', file=sys.stderr)
        print(errors, file=sys.stderr, flush=True)

        if allow_fail:
            return False, errors
        else:
            sys.exit(1)

    return True, None


def decode(x):
    try:
        s = x.decode()
    except:
        return x
    return s



class Error (Exception): pass

class Link:
    def __init__(self, sam, sam_reader, ref_lengths, s=None):
        # s meant for writing tests. Construct object from a string (same format as __str__())
        if s is not None:
            l = s.split('\t')
            assert len(l) == 8
            self.refnames = [l[0], l[4]]
            self.lengths = [None if l[i] == '.' else int(l[i]) for i in [1,5]]
            self.dirs = [l[2], l[6]]
            self.pos = [None if l[i] == '.' else int(l[i]) for i in [3,7]]
            return

        if sam.is_unmapped or sam.mate_is_unmapped or sam.reference_id == sam.next_reference_id:
            raise Error('Cannot create link from sam:\n' + str(sam))

        self.refnames = [sam_reader.getrname(sam.reference_id), sam_reader.getrname(sam.next_reference_id)]
        self.dirs = ['L' if sam.is_reverse else 'R', 'L' if sam.mate_is_reverse else 'R']
        position = sam.reference_end - 1 if sam.is_reverse else sam.reference_start
        self.pos = [position, None]
        self.lengths = [ref_lengths[x] for x in self.refnames]
        
        if not sam.is_read1:
            self._swap()


    def __eq__(self, other):
        return type(other) is type(self) and self.__dict__ == other.__dict__


    def __lt__(self, other):
        if self.refnames[0] < other.refnames[0]:
            return True
        elif self.refnames[0] > other.refnames[0]:
            return False
        elif self.pos[0] < other.pos[0]:
            return True
        elif self.pos[0] == other.pos[0]:
            return self.pos[1] < other.pos[1]
        else:
            return False


    def __str__(self):
        return '\t'.join([
           self.refnames[0], 
           str(self.lengths[0]),
           self.dirs[0],
           str(self.pos[0]) if self.pos[0] is not None else '.',
           self.refnames[1], 
           str(self.lengths[1]),
           self.dirs[1],
           str(self.pos[1]) if self.pos[1] is not None else '.'
        ])


    def _swap(self):
        for l in [self.refnames, self.dirs, self.pos, self.lengths]:
            l.reverse()


             

    def sort(self):
        if self.refnames[1] < self.refnames[0]:
            self._swap()


    def _distance_to_contig_end(self, one_or_two):
        one_or_two -= 1
        if self.dirs[one_or_two] == 'L' and self.pos[one_or_two] is not None:
            return self.pos[one_or_two]
        elif self.dirs[one_or_two] == 'R' and self.pos[one_or_two] is not None:
            return self.lengths[one_or_two] - self.pos[one_or_two] - 1
        else:
            raise Error('Error in _distance_to_contig_end(' + str(one_or_two + 1) + ') for link:\n' + str(self))
        

    def merge(self, other):
        '''Merge another link into this one. Expected that each link was created from each mate from a pair. We only know both distances to contig ends when we have read info from both mappings in a BAM file. All other info should be the same.'''
        assert self.refnames == other.refnames
        assert self.dirs == other.dirs
        assert self.lengths == other.lengths

        for i in range(2):
            if self.pos[i] is None:
                if other.pos[i] is None:
                   raise Error('Error merging these two links:\n' + str(self) + '\n' + str(other))
                self.pos[i] = other.pos[i]
            else:
                if other.pos[i] is not None:
                   raise Error('Error merging these two links:\n' + str(self) + '\n' + str(other))


    def _distance_to_contig_ends(self):
        return self._distance_to_contig_end(1), self._distance_to_contig_end(2)


    def insert_size(self):
        '''Returns insert size, defined as distance from outer edges of reads (and assumes gap length of zero)'''
        try:
            distances = self._distance_to_contig_ends()
        except:
            raise Error('Error getting insert size from Link:\n' + str(self))

        return sum(distances)



__all__ = [
    'bam_parse',
    'cdhit',
    'cluster',
    'clusters',
    'common',
    'external_progs',
    'faidx',
    'flag',
    'histogram',
    'link',
    'mapping',
    'refcheck',
    'scaffold_graph',
    'summary',
    'tasks',
]

from ariba import *



import os
import pyfastaq

class Error (Exception): pass


class Checker:
    def __init__(self, infile, min_length=1, max_length=10000):
        self.infile = os.path.abspath(infile)
        if not os.path.exists(self.infile):
            raise Error('File not found: "' + self.infile + '". Cannot continue')

        self.min_length = min_length
        self.max_length = max_length


    def check(self, error_code_on_exit=None):
        file_reader = pyfastaq.sequences.file_reader(self.infile)

        for seq in file_reader:
            if not seq.looks_like_gene():
                return False, 'Not a gene', seq
            elif len(seq) < self.min_length:
                return False, 'Too short', seq
            elif len(seq) > self.max_length:
                return False, 'Too long', seq
           
        return True, None, None


    def fix(self, outprefix):
        file_reader = pyfastaq.sequences.file_reader(self.infile)
        old2new_out = outprefix + '.rename'
        fasta_out = outprefix + '.fa'
        bad_seqs_out = outprefix + '.removed.fa'
        log_out = outprefix + '.log'
        names = {}
        old2new_out_fh = pyfastaq.utils.open_file_write(old2new_out)
        fasta_out_fh = pyfastaq.utils.open_file_write(fasta_out)
        bad_seqs_out_fh = pyfastaq.utils.open_file_write(bad_seqs_out)
        log_out_fh = pyfastaq.utils.open_file_write(log_out)

        for seq in file_reader:
            seq.seq = seq.seq.upper()
            if len(seq) < self.min_length:
                print(seq.id, 'Too short. Skipping', sep='\t', file=log_out_fh)
                print(seq, file=bad_seqs_out_fh)
                continue
            elif len(seq) > self.max_length:
                print(seq.id, 'Too long. Skipping', sep='\t', file=log_out_fh)
                print(seq, file=bad_seqs_out_fh)
                continue


            if not seq.looks_like_gene():
                seq.revcomp()
                if seq.looks_like_gene():
                    print(seq.id, 'Reverse complemented', sep='\t', file=log_out_fh)
                else:
                    print(seq.id, 'Does not look like a gene. Skipping', sep='\t', file=log_out_fh)
                    seq.revcomp()
                    print(seq, file=bad_seqs_out_fh)
                    continue
                
            original_id = seq.id
            # replace unwanted characters with underscores
            to_replace = ' '
            seq.id = seq.id.translate(str.maketrans(to_replace, '_' *  len(to_replace)))

            if seq.id in names:
                names[seq.id] += 1
                seq.id += '.' + str(names[seq.id])
            else:
                names[seq.id] = 1

            print(original_id, seq.id, sep='\t', file=old2new_out_fh)
            print(seq, file=fasta_out_fh)

        pyfastaq.utils.close(fasta_out_fh)
        pyfastaq.utils.close(bad_seqs_out_fh)
        pyfastaq.utils.close(log_out_fh)
        pyfastaq.utils.close(old2new_out_fh)



import argparse
import ariba

def run():
    parser = argparse.ArgumentParser(
        description = 'Make a summry of ARIBA report files',
        usage = 'ariba summary [options] <outfile> [infiles]',
        epilog = 'Files must be listed after the output file and/or the option --fofn must be used. If both used, all files in the filename specified by --fofn AND the files listed after the output file will be used as input')
    parser.add_argument('-f', '--fofn', help='File of filenames of ariba reports to be summarised. Must b used if no input files listed after the outfile', metavar='FILENAME')
    parser.add_argument('--min_id', type=float, help='Minimum percent identity cutoff to count as assembled [%(default)s]', default=90, metavar='FLOAT')
    parser.add_argument('outfile', help='Name of output file. If file ends with ".xls", then an excel spreadsheet is written. Otherwise a tsv file is written')
    parser.add_argument('infiles', nargs='*', help='Files to be summarised')
    options = parser.parse_args()
    if len(options.infiles) == 0:
        options.infiles = None

    s = ariba.summary.Summary(
        options.outfile,
        fofn=options.fofn,
        filenames=options.infiles,
        min_id=options.min_id
    ) 
    s.run()



import os
import copy
from operator import itemgetter
import sys
import shutil
import pysam
import operator
import pyfastaq
import pymummer
from ariba import common, mapping, bam_parse, flag, faidx

class Error (Exception): pass


class Cluster:
    def __init__(self,
      root_dir,
      name,
      assembly_kmer=0,
      assembler='velvet',
      max_insert=1000,
      min_scaff_depth=10,
      nucmer_min_id=90,
      nucmer_min_len=50,
      nucmer_breaklen=50,
      sspace_k=20,
      reads_insert=500,
      sspace_sd=0.4,
      threads=1,
      bcf_min_dp=10,
      bcf_min_dv=5,
      bcf_min_dv_over_dp=0.3,
      bcf_min_qual=20,
      assembled_threshold=0.95,
      unique_threshold=0.03,
      verbose=False,
      bcftools_exe='bcftools',
      gapfiller_exe='GapFiller.pl',
      samtools_exe='samtools',
      bowtie2_exe='bowtie2',
      bowtie2_preset='very-sensitive-local',
      smalt_exe='smalt',
      spades_exe='spades.py',
      sspace_exe='SSPACE_Basic_v2.0.pl',
      velvet_exe='velvet', # prefix of velvet{g,h}
      spades_other=None,
      clean=1,
    ):

        self.root_dir = os.path.abspath(root_dir)
        if not os.path.exists(self.root_dir):
            raise Error('Directory ' + self.root_dir + ' not found. Cannot continue')

        self.name = name
        self.reads1 = os.path.join(self.root_dir, 'reads_1.fq')
        self.reads2 = os.path.join(self.root_dir, 'reads_2.fq')
        self.gene_fa = os.path.join(self.root_dir, 'gene.fa')
        self.genes_fa = os.path.join(self.root_dir, 'genes.fa')
        self.gene_bam = os.path.join(self.root_dir, 'gene.reads_mapped.bam')

        for fname in [self.reads1, self.reads2, self.genes_fa]:
            if not os.path.exists(fname):
                raise Error('File ' + fname + ' not found. Cannot continue')


        self.max_insert = max_insert
        self.min_scaff_depth = min_scaff_depth

        self.nucmer_min_id = nucmer_min_id
        self.nucmer_min_len = nucmer_min_len
        self.nucmer_breaklen = nucmer_breaklen
        self.assembly_vs_gene_coords = os.path.join(self.root_dir, 'assembly_vs_gene.coords')

        self.bcf_min_dp = bcf_min_dp
        self.bcf_min_dv = bcf_min_dv
        self.bcf_min_dv_over_dp = bcf_min_dv_over_dp
        self.bcf_min_qual = bcf_min_qual

        self._set_assembly_kmer(assembly_kmer)
        self.assembler = assembler
        assert self.assembler in ['velvet', 'spades']
        self.spades_exe = spades_exe
        self.spades_other = spades_other

        self.bcftools_exe = bcftools_exe

        self.sspace_exe = shutil.which(sspace_exe)
        if self.sspace_exe is None:
            self.gapfiller_exe = None
        else:
            self.sspace_exe = os.path.realpath(self.sspace_exe) # otherwise sspace dies loading packages
            self.gapfiller_exe = shutil.which(gapfiller_exe)
            if self.gapfiller_exe is not None:
                self.gapfiller_exe = os.path.realpath(self.gapfiller_exe) # otherwise gapfiller dies loading packages

        self.samtools_exe = samtools_exe
        self.smalt_exe = smalt_exe
        self.bowtie2_exe = bowtie2_exe
        self.bowtie2_preset = bowtie2_preset

        if self.assembler == 'velvet':
            self.velveth = velvet_exe + 'h'
            self.velvetg = velvet_exe + 'g'

        self.sspace_k = sspace_k
        self.reads_insert = reads_insert
        self.sspace_sd = sspace_sd

        self.threads = threads
        self.verbose = verbose
        self.assembled_threshold = assembled_threshold
        self.unique_threshold = unique_threshold
        self.status_flag = flag.Flag()
        self.flag_file = os.path.join(self.root_dir, 'flag')
        self.clean = clean

        self.assembly_dir = os.path.join(self.root_dir, 'Assembly')
        try:
            os.mkdir(self.assembly_dir)
        except:
            raise Error('Error mkdir ' + self.assembly_dir)
        self.assembler_dir = os.path.join(self.assembly_dir, 'Assemble')
        self.assembly_contigs = os.path.join(self.assembly_dir, 'contigs.fa')
        self.scaffold_dir = os.path.join(self.assembly_dir, 'Scaffold')
        self.scaffolder_scaffolds = os.path.join(self.assembly_dir, 'scaffolds.fa')
        self.gapfill_dir = os.path.join(self.assembly_dir, 'Gapfill')
        self.gapfilled_scaffolds = os.path.join(self.assembly_dir, 'scaffolds.gapfilled.fa')
        self.final_assembly_fa = os.path.join(self.root_dir, 'assembly.fa')
        self.final_assembly_bam = os.path.join(self.root_dir, 'assembly.reads_mapped.bam')
        self.final_assembly_read_depths = os.path.join(self.root_dir, 'assembly.reads_mapped.bam.read_depths.gz')
        self.final_assembly_vcf = os.path.join(self.root_dir, 'assembly.reads_mapped.bam.vcf')
        self.final_assembly = {}
        self.mummer_variants = {}
        self.variant_depths = {}
        self.percent_identities = {}


    def _get_read_counts(self):
        count1 = pyfastaq.tasks.count_sequences(self.reads1)
        count2 = pyfastaq.tasks.count_sequences(self.reads2)
        if count1 == count2:
            return count1 + count2
        else:
            raise Error('Different number of fwd/rev reads in cluster ' + self.name + '! Cannot continue')


    def _get_total_alignment_score(self, gene_name):
        tmp_bam = os.path.join(self.root_dir, 'tmp.get_total_alignment_score.bam')
        assert not os.path.exists(tmp_bam)
        tmp_fa = os.path.join(self.root_dir, 'tmp.get_total_alignment_score.ref.fa')
        assert not os.path.exists(tmp_fa)
        faidx.write_fa_subset([gene_name], self.genes_fa, tmp_fa, samtools_exe=self.samtools_exe, verbose=self.verbose)
        mapping.run_bowtie2(
            self.reads1,
            self.reads2,
            tmp_fa,
            tmp_bam[:-4],
            threads=self.threads,
            samtools=self.samtools_exe,
            bowtie2=self.bowtie2_exe,
            bowtie2_preset=self.bowtie2_preset,
            verbose=self.verbose,
        )

        score = mapping.get_total_alignment_score(tmp_bam)
        os.unlink(tmp_bam)
        os.unlink(tmp_fa)
        os.unlink(tmp_fa + '.fai')
        return score


    def _get_best_gene_by_alignment_score(self):
        cluster_size = pyfastaq.tasks.count_sequences(self.genes_fa)
        if cluster_size == 1:
            seqs = {}
            pyfastaq.tasks.file_to_dict(self.genes_fa, seqs)
            assert len(seqs) == 1
            gene_name = list(seqs.values())[0].id
            if self.verbose:
                print('No need to choose gene for this cluster because only has one gene:', gene_name)
            return gene_name

        if self.verbose:
            print('\nChoosing best gene from cluster of', cluster_size, 'genes...')
        file_reader = pyfastaq.sequences.file_reader(self.genes_fa)
        best_score = 0
        best_gene_name = None
        for seq in file_reader:
            score = self._get_total_alignment_score(seq.id)
            if self.verbose:
                print('Total alignment score for gene', seq.id, 'is', score)
            if score > best_score:
                best_score = score
                best_gene_name = seq.id

        if self.verbose:
            print('Best gene is', best_gene_name, 'with total alignment score of', best_score)
            print()

        return best_gene_name


    def _choose_best_gene(self):
        gene_name = self._get_best_gene_by_alignment_score()
        if gene_name is None:
            return None
        faidx.write_fa_subset([gene_name], self.genes_fa, self.gene_fa, samtools_exe=self.samtools_exe, verbose=self.verbose)
        seqs = {}
        pyfastaq.tasks.file_to_dict(self.gene_fa, seqs)
        assert len(seqs) == 1
        return list(seqs.values())[0]


    def _set_assembly_kmer(self, k):
        '''If the kmer not given, uses 2/3 of the mean read length (using first 1000 forward and first 1000 reverse reads)'''
        if k == 0:
            read_length1 = pyfastaq.tasks.mean_length(self.reads1, limit=1000)
            read_length2 = pyfastaq.tasks.mean_length(self.reads2, limit=1000)
            self.assembly_kmer = round( (read_length1 + read_length2) / 3)
            if self.assembly_kmer % 2 == 0:
                self.assembly_kmer += 1
        else:
            self.assembly_kmer = k


    def _assemble_with_velvet(self):
        # map reads to reference gene to make BAM input to velvet columbus
        mapping.run_bowtie2(
            self.reads1,
            self.reads2,
            self.gene_fa,
            self.gene_bam[:-4],
            threads=self.threads,
            sort=True,
            samtools=self.samtools_exe,
            bowtie2=self.bowtie2_exe,
            bowtie2_preset=self.bowtie2_preset,
            verbose=self.verbose,
        )

        cmd = ' '.join([
            self.velveth,
            self.assembler_dir,
            str(self.assembly_kmer),
            '-reference', self.gene_fa,
            '-shortPaired -bam', self.gene_bam[:-4] + '.unsorted.bam'
        ])

        cwd = os.getcwd()
        os.chdir(self.assembly_dir)
        velvet_contigs = os.path.join(os.path.split(self.assembler_dir)[1], 'contigs.fa')

        self.velveth_ok, err = common.syscall(cmd, verbose=self.verbose, allow_fail=True)
        if not self.velveth_ok:
            with open('velveth_errors', 'w') as f:
                print(err, file=f)
                f.close()
            self.status_flag.add('assembly_fail')
            os.chdir(cwd)
            return

        cmd = ' '.join([
            self.velvetg,
            self.assembler_dir,
            '-ins_length', str(int(self.reads_insert)),
            '-scaffolding no',
            '-exp_cov auto',
            '-very_clean yes',
            '-cov_cutoff auto',
        ])

        self.assembled_ok, err = common.syscall(cmd, verbose=self.verbose, allow_fail=True)
        if self.assembled_ok:
            os.symlink(velvet_contigs, os.path.basename(self.assembly_contigs))
        else:
            with open('velvetg_errors', 'w') as f:
                print(err, file=f)
                f.close()
            self.status_flag.add('assembly_fail')

        os.chdir(cwd)


    def _assemble_with_spades(self, unittest=False):
        cmd = ' '.join([
            self.spades_exe,
            '-1', self.reads1,
            '-2', self.reads2,
            '-o', self.assembler_dir,
            '-k', str(self.assembly_kmer),
            '--threads', str(self.threads),
            '--untrusted-contigs', self.gene_fa,
        ])
        if self.spades_other is not None:
            cmd += ' ' + self.spades_other

        cwd = os.getcwd()
        os.chdir(self.assembly_dir)
        spades_contigs = os.path.join(os.path.split(self.assembler_dir)[1], 'scaffolds.fasta')

        if unittest:
            os.mkdir(self.assembler_dir)
            open(spades_contigs, 'w').close()
            self.assembled_ok = True
        else:
            self.assembled_ok, err = common.syscall(cmd, verbose=self.verbose, allow_fail=True)
        if self.assembled_ok:
            os.symlink(spades_contigs, os.path.basename(self.assembly_contigs))
        else:
            with open('spades_errors', 'w') as f:
                print(err, file=f)
            f.close()
            self.status_flag.add('assembly_fail')

        os.chdir(cwd)


    def _scaffold_with_sspace(self):
        if not os.path.exists(self.assembly_contigs):
            raise Error('Cannot scaffold because contigs file not found: ' + self.assembly_contigs)

        try:
            os.mkdir(self.scaffold_dir)
        except:
            raise Error('Error mkdir '+  self.scaffold_dir)

        cwd = os.getcwd()

        if self.sspace_exe is None:
            os.chdir(self.assembly_dir)
            os.symlink(os.path.basename(self.assembly_contigs), os.path.basename(self.scaffolder_scaffolds))
            os.chdir(cwd)
            return

        os.chdir(self.scaffold_dir)
        lib_file = 'lib'
        with open(lib_file, 'w') as f:
            print('LIB', self.reads1, self.reads2, int(self.reads_insert), self.sspace_sd, 'FR', file=f)

        cmd = ' '.join([
            'perl', self.sspace_exe,
            '-k', str(self.sspace_k),
            '-l', lib_file,
            '-s', self.assembly_contigs
        ])

        sspace_scaffolds = os.path.abspath('standard_output.final.scaffolds.fasta')
        common.syscall(cmd, verbose=self.verbose)
        os.chdir(self.assembly_dir)
        os.symlink(os.path.relpath(sspace_scaffolds), os.path.basename(self.scaffolder_scaffolds))
        os.chdir(cwd)


    def _has_gaps_to_fill(self, filename):
        seq_reader = pyfastaq.sequences.file_reader(filename)
        for seq in seq_reader:
            if 'n' in seq.seq or 'N' in seq.seq:
                return True
        return False


    def _gap_fill_with_gapfiller(self):
        if not os.path.exists(self.scaffolder_scaffolds):
            raise Error('Cannot gap fill because scaffolds file not found: ' + self.scaffolder_scaffolds)


        cwd = os.getcwd()

        if self.gapfiller_exe is None or not self._has_gaps_to_fill(self.scaffolder_scaffolds):
            self._rename_scaffolds(self.scaffolder_scaffolds, self.gapfilled_scaffolds)
            return

        try:
            os.mkdir(self.gapfill_dir)
        except:
            raise Error('Error mkdir '+  self.gapfill_dir)

        os.chdir(self.gapfill_dir)
        lib_file = 'lib'
        with open(lib_file, 'w') as f:
            print('LIB', 'bwa', self.reads1, self.reads2, self.reads_insert, self.sspace_sd, 'FR', file=f)

        cmd = ' '.join([
            'perl', self.gapfiller_exe,
            '-l', lib_file,
            '-s', self.scaffolder_scaffolds
        ])

        gapfilled_scaffolds = os.path.join(self.gapfill_dir, 'standard_output', 'standard_output.gapfilled.final.fa')
        common.syscall(cmd, verbose=self.verbose)
        self._rename_scaffolds(gapfilled_scaffolds, self.gapfilled_scaffolds)
        os.chdir(cwd)


    def _rename_scaffolds(self, infile, outfile):
        freader = pyfastaq.sequences.file_reader(infile)
        f_out = pyfastaq.utils.open_file_write(outfile)
        i = 1
        for scaff in freader:
            scaff.id = self.gene.id + '.scaffold.' + str(i)
            i += 1
            print(scaff, file=f_out)
        pyfastaq.utils.close(f_out)


    def _run_nucmer(self, qry, outfile, show_snps=False):
        pymummer.nucmer.Runner(
            self.gene_fa,
            qry,
            outfile,
            min_id=self.nucmer_min_id,
            min_length=self.nucmer_min_len,
            breaklen=self.nucmer_breaklen,
            show_snps=show_snps
        ).run()


    def _fix_contig_orientation(self):
        if not os.path.exists(self.gapfilled_scaffolds):
            raise Error('Cannot fix orientation of assembly contigs because file not found: ' + self.gapfilled_scaffolds)

        tmp_coords = os.path.join(self.root_dir, 'tmp.coords')
        self._run_nucmer(self.gapfilled_scaffolds, tmp_coords)

        to_revcomp = set()
        not_revcomp = set()
        file_reader = pymummer.coords_file.reader(tmp_coords)
        for hit in file_reader:
            if hit.on_same_strand():
                not_revcomp.add(hit.qry_name)
            else:
                to_revcomp.add(hit.qry_name)

        os.unlink(tmp_coords)
        in_both = to_revcomp.intersection(not_revcomp)
        for name in in_both:
            print('WARNING: hits to both strands of gene for scaffold. Interpretation of any variants cannot be trusted for this scaffold:', name, file=sys.stderr)
            to_revcomp.remove(name)
            self.status_flag.add('hit_both_strands')

        f = pyfastaq.utils.open_file_write(self.final_assembly_fa)
        seq_reader = pyfastaq.sequences.file_reader(self.gapfilled_scaffolds)
        for seq in seq_reader:
            if seq.id in to_revcomp:
                seq.revcomp()
            print(seq, file=f)
        pyfastaq.utils.close(f)


    def _load_final_contigs(self):
        if not os.path.exists(self.final_assembly_fa):
            raise Error('Cannot load final assembled contigs because file not found:' + self.final_assembly_fa)

        self.final_assembly = {}
        pyfastaq.tasks.file_to_dict(self.final_assembly_fa, self.final_assembly)


    def _parse_assembly_bam(self):
        if not os.path.exists(self.final_assembly_bam):
            raise Error('File not found: ' + self.final_assembly_bam)

        bam_parser = bam_parse.Parser(self.final_assembly_bam, self.final_assembly)
        bam_parser.parse()
        bam_parser.write_files(self.final_assembly_bam)
        if not bam_parser.scaff_graph_is_consistent(self.min_scaff_depth, self.max_insert):
            self.status_flag.add('scaffold_graph_bad')


    def _parse_assembly_vs_gene_coords(self):
        file_reader = pymummer.coords_file.reader(self.assembly_vs_gene_coords)
        self.nucmer_hits = {}
        for hit in file_reader:
            assert hit.ref_name == self.gene.id
            contig = hit.qry_name
            if contig not in self.nucmer_hits:
                self.nucmer_hits[contig] = []
            self.nucmer_hits[contig].append(copy.copy(hit))


    def _nucmer_hits_to_percent_identity(self):
        self.percent_identities = {}
        for contig in self.nucmer_hits:
            product_sum = 0
            length_sum = 0
            for hit in self.nucmer_hits[contig]:
                product_sum += hit.hit_length_qry * hit.percent_identity
                length_sum += hit.hit_length_qry
            assert length_sum > 0
            self.percent_identities[contig] = round(product_sum / length_sum, 2)


    def _nucmer_hits_to_scaff_coords(self):
        coords = {}
        for l in self.nucmer_hits.values():
            for hit in l:
                if hit.qry_name not in coords:
                    coords[hit.qry_name] = []
                coords[hit.qry_name].append(hit.qry_coords())

        for scaff in coords:
            pyfastaq.intervals.merge_overlapping_in_list(coords[scaff])

        return coords


    def _nucmer_hits_to_ref_coords(self, contig=None):
        coords = []
        if contig is None:
            keys = list(self.nucmer_hits.keys())
        else:
            keys = [contig]

        for key in keys:
            coords += [hit.ref_coords() for hit in self.nucmer_hits[key]]
        coords.sort()
        return coords


    def _nucmer_hits_to_gene_cov_per_contig(self):
        cov = {}
        for contig in self.nucmer_hits:
            coords = self._nucmer_hits_to_ref_coords(contig)
            pyfastaq.intervals.merge_overlapping_in_list(coords)
            cov[contig] = pyfastaq.intervals.length_sum_from_list(coords)
        return cov


    def _whole_gene_covered_by_nucmer_hits(self):
        covered = self._nucmer_hits_to_ref_coords()
        pyfastaq.intervals.merge_overlapping_in_list(covered)
        return pyfastaq.intervals.length_sum_from_list(covered) / len(self.gene) >= self.assembled_threshold


    def _gene_coverage_unique(self):
        covered = self._nucmer_hits_to_ref_coords()
        covered.sort()
        if len(covered) <= 1:
            return True

        coverage = {}
        for i in covered:
            for j in range(i.start, i.end + 1):
                coverage[j] = coverage.get(j, 0) + 1

        bases_depth_at_least_two = len([1 for x in coverage.values() if x > 1])
        return bases_depth_at_least_two / len(self.gene) <= self.unique_threshold


    def _gene_covered_by_complete_contig_with_orf(self):
        for l in self.nucmer_hits.values():
            for hit in l:
                if hit.hit_length_ref == len(self.gene):
                    start = min(hit.qry_start, hit.qry_end)
                    end = max(hit.qry_start, hit.qry_end)
                    assembled_gene = pyfastaq.sequences.Fasta('x', self.final_assembly[hit.qry_name][start:end+1])
                    if (hit.ref_start < hit.ref_end) != (hit.qry_start < hit.qry_end):
                        assembled_gene.revcomp()
                    assembled_gene_aa = assembled_gene.translate()
                    orfs = assembled_gene.orfs()
                    if len(orfs) == 0:
                        continue

                    max_orf = orfs[0]
                    for o in orfs:
                        if len(o) > len(max_orf):
                            max_orf = o

                    if len(max_orf) == len(assembled_gene):
                        return True
        return False


    def _gene_covered_by_at_least_one_full_length_contig(self):
        for l in self.nucmer_hits.values():
            for hit in l:
                if len(hit.ref_coords()) == len(self.gene):
                    return True
        return False


    def _update_flag_from_nucmer_file(self):
        if self._whole_gene_covered_by_nucmer_hits():
            self.status_flag.add('gene_assembled')

        if self._gene_covered_by_at_least_one_full_length_contig():
            self.status_flag.add('gene_assembled_into_one_contig')

        if not self._gene_coverage_unique():
            self.status_flag.add('gene_region_assembled_twice')

        if self._gene_covered_by_complete_contig_with_orf():
            self.status_flag.add('complete_orf')

        if len(self.nucmer_hits) == 1:
            self.status_flag.add('unique_contig')


    def _get_mummer_variants(self):
        snp_file = self.assembly_vs_gene_coords + '.snps'
        if not os.path.exists(snp_file):
            raise Error('File not found ' + snp_file)
        variants = pymummer.snp_file.get_all_variants(snp_file)
        self.mummer_variants = {}

        if len(variants) == 0:
            return

        variants.sort(key=operator.attrgetter('qry_name'))
        variants.sort(key=operator.attrgetter('ref_start'))

        for v in variants:
            if v.qry_name not in self.mummer_variants:
                self.mummer_variants[v.qry_name] = []
            self.mummer_variants[v.qry_name].append(v)

        for contig in self.mummer_variants:
            l = self.mummer_variants[contig]
            if len(l) > 1:
                new_l = [[l[0]]]
                previous_codon_start = self._get_codon_start(0, l[0].ref_start)
                for variant in l[1:]:
                    codon_start = self._get_codon_start(0, variant.ref_start)
                    if codon_start == previous_codon_start:
                        new_l[-1].append(variant)
                    else:
                        new_l.append([variant])
                        previous_codon_start = codon_start
                self.mummer_variants[contig] = new_l
            else:
                self.mummer_variants[contig] = [l]


    def _filter_mummer_variants(self):
        if len(self.mummer_variants) == 0:
            return

        for contig in self.mummer_variants:
            variants = self.mummer_variants[contig]
            for i in range(len(variants)):
                t = self._get_variant_effect(variants[i])
                if t is not None and t[0] in ['TRUNC', 'FSHIFT']:
                    break
            self.mummer_variants[contig] = variants[:i+1]


    def _get_codon_start(self, gene_start, position):
        assert position >= gene_start
        while  (position - gene_start) % 3 != 0:
            position -= 1
        return position


    def _get_variant_effect(self, variants):
        if len(variants) == 0:
            return None

        var_types = [x.var_type for x in variants]
        if len(set(var_types)) != 1:
            return None

        var_type = var_types[0]

        assert set([x.ref_name for x in variants]) == set([self.gene.id])
        codon_starts = [self._get_codon_start(0, x.ref_start) for x in variants]
        assert len(set(codon_starts)) == 1
        codon_start = codon_starts[0]
        aa_start = codon_start // 3
        ref_codon = pyfastaq.sequences.Fasta('codon', self.gene[codon_start:codon_start+3])
        ref_aa = ref_codon.translate()

        if var_type == pymummer.variant.SNP:
            new_codon = list(ref_codon.seq)
            for v in variants:
                new_codon[v.ref_start - codon_start] = v.qry_base
            new_codon = pyfastaq.sequences.Fasta('new', ''.join(new_codon))
            qry_aa = new_codon.translate()

            if ref_aa.seq == qry_aa.seq:
                return ('SYN', '.')
            elif qry_aa.seq == '*':
                return ('TRUNC', ref_aa.seq + str(aa_start + 1) + 'trunc')
            else:
                return ('NONSYN', ref_aa.seq + str(aa_start + 1) + qry_aa.seq)
        elif var_type in [pymummer.variant.INS, pymummer.variant.DEL]:
            if len(variants) > 1:
                print('More than one indel in same codon not yet implemented!', self.gene.id, file=sys.stderr)
                return None

            var = variants[0]

            if var_type == pymummer.variant.INS:
                new_seq = pyfastaq.sequences.Fasta('seq', var.qry_base)
            else:
                new_seq = pyfastaq.sequences.Fasta('seq', var.ref_base)

            if len(new_seq) % 3 != 0:
                return ('FSHIFT', ref_aa.seq + str(aa_start + 1) + 'fs')

            new_seq_aa = new_seq.translate()
            if '*' in new_seq_aa.seq:
                return ('TRUNC', ref_aa.seq + str(aa_start + 1) + 'trunc')
            elif var_type == pymummer.variant.INS:
                ref_codon_after_ins = pyfastaq.sequences.Fasta('codon', self.gene[codon_start+3:codon_start+6])
                aa_after_ins = ref_codon_after_ins.translate()
                return ('INS', ref_aa.seq + str(aa_start + 1) + '_' + aa_after_ins.seq + str(aa_start + 2) + 'ins' + new_seq_aa.seq )
            else:
                if len(new_seq) == 3:
                    return ('DEL', ref_aa.seq + str(aa_start + 1) + 'del')
                else:
                    assert len(new_seq) % 3 == 0
                    new_aa = new_seq.translate()
                    ref_codon_after_ins = pyfastaq.sequences.Fasta('codon', self.gene[codon_start+3:codon_start+6])
                    aa_after_ins = ref_codon_after_ins.translate()
                    return ('DEL', ref_aa.seq + str(aa_start + 1)+ '_' + aa_after_ins.seq + str(aa_start + 2) + 'del')

        else:
            return ('UNKNOWN', '.')


    def _make_assembly_vcf(self):
        tmp_vcf = self.final_assembly_vcf + '.tmp'
        cmd = ' '.join([
            self.samtools_exe, 'mpileup',
            '-t INFO/DPR,DV',
            '-A',
            '-f', self.final_assembly_fa,
            '-u',
            '-v',
            self.final_assembly_bam,
            '>',
            tmp_vcf
        ])

        common.syscall(cmd, verbose=self.verbose)

        cmd = ' '.join([
            self.bcftools_exe, 'call -m',
            tmp_vcf,
            '|',
            self.bcftools_exe, 'query',
            r'''-f '%CHROM\t%POS\t%REF\t%ALT\t%DP\t%DPR]\n' ''',
            '>',
            self.final_assembly_read_depths + '.tmp'
        ])

        common.syscall(cmd, verbose=self.verbose)
        pysam.tabix_compress(self.final_assembly_read_depths + '.tmp', self.final_assembly_read_depths)
        pysam.tabix_index(self.final_assembly_read_depths, seq_col=0, start_col=1, end_col=1)
        os.unlink(self.final_assembly_read_depths + '.tmp')

        cmd = ' '.join([
            self.bcftools_exe, 'call -m -v',
            tmp_vcf,
            '|',
            self.bcftools_exe, 'filter',
            '-i', '"MIN(DP)>=' + str(self.bcf_min_dp),
                  ' & MIN(DV)>=' + str(self.bcf_min_dv),
                  ' & MIN(DV/DP)>=' + str(self.bcf_min_dv_over_dp),
                  ' & QUAL >=', str(self.bcf_min_qual), '"',
            '-o', self.final_assembly_vcf
        ])

        common.syscall(cmd, verbose=self.verbose)
        os.unlink(tmp_vcf)


    def _get_assembly_read_depths(self, ref, position):
        '''Returns total read depth and depth of reads supporting alternative (if present)'''
        assert os.path.exists(self.final_assembly_read_depths)
        assert os.path.exists(self.final_assembly_read_depths + '.tbi')
        tbx = pysam.TabixFile(self.final_assembly_read_depths)
        try:
            rows = [x for x in tbx.fetch(ref, position, position + 1)]
        except:
            return None

        if len(rows) > 1: # which happens with indels, mutiple lines for same base of reference
            test_rows = [x for x in rows if x.rstrip().split()[3] != '.']
            if len(test_rows) != 1:
                rows = [rows[-1]]
            else:
                rows = test_rows

        if len(rows) == 1:
            r, p, ref_base, alt_base, ref_counts, alt_counts = rows[0].rstrip().split()
            return ref_base, alt_base, int(ref_counts), alt_counts
        else:
            return None


    def _get_samtools_variant_positions(self):
        if not os.path.exists(self.final_assembly_vcf):
            return []
        f = pyfastaq.utils.open_file_read(self.final_assembly_vcf)
        positions = [l.rstrip().split('\t')[0:2] for l in f if not l.startswith('#')]
        positions = [(t[0], int(t[1]) - 1) for t in positions]
        pyfastaq.utils.close(f)
        return positions


    def _get_samtools_variants(self, positions=None):
        if positions is None:
            positions = self._get_samtools_variant_positions()
        variants = {}
        if len(positions) == 0:
            return variants
        if not (os.path.exists(self.final_assembly_vcf) and os.path.exists(self.final_assembly_read_depths)):
            return variants
        for t in positions:
            name, pos = t[0], t[1]
            depths = self._get_assembly_read_depths(name, pos)
            if depths is None:
                raise Error('Error getting read depths for sequence ' + name + ' at position ' + t[1])
            if name not in variants:
                variants[name] = {}
            variants[name][t[1]] = depths
        return variants


    def _get_vcf_variant_counts(self):
        scaff_coords = self._nucmer_hits_to_scaff_coords()
        self.vcf_variant_counts = {}
        f = pyfastaq.utils.open_file_read(self.final_assembly_vcf)
        for line in f:
            if line.startswith('#'):
                continue

            data = line.rstrip().split('\t')
            scaff = data[0]

            if scaff in scaff_coords:
                position = int(data[1]) - 1
                i = pyfastaq.intervals.Interval(position, position)
                intersects = len([x for x in scaff_coords[scaff] if x.intersects(i)]) > 0
                if intersects:
                    self.vcf_variant_counts[scaff] = self.vcf_variant_counts.get(scaff, 0) + 1

        pyfastaq.utils.close(f)
        total = sum(list(self.vcf_variant_counts.values()))
        if total >= 1:
            self.status_flag.add('variants_suggest_collapsed_repeat')


    def _make_report_lines(self):
        self.report_lines = []
        total_reads = self._get_read_counts()

        if not self.assembled_ok:
            gene_name = 'NA' if self.gene is None else self.gene.id
            gene_length = '.' if self.gene is None else len(self.gene)
            self.report_lines.append([
                    gene_name,
                    self.status_flag.to_number(),
                    total_reads,
                    self.name,
                    gene_length,
                    '.',
                    '.',
                  ] + \
                  ['.'] * 14
            )
            return

        cov_per_contig = self._nucmer_hits_to_gene_cov_per_contig()
        samtools_variants = self._get_samtools_variants()


        for contig in self.mummer_variants:
            for variants in self.mummer_variants[contig]:
                t = self._get_variant_effect(variants)
                if t is not None:
                    effect, new_bases = t
                    for v in variants:
                        depths = self._get_assembly_read_depths(contig, v.qry_start)
                        if depths is None:
                            # this happens with low coverage contigs. It can get assembled, but
                            # there are some bases that do not have reads mapped to them.
                            # If mummer called a variant at one of these, then we're looking
                            # for read dpeth where there is none.
                            print('Warning: could not get read depth info on contig "' + contig + '" at position ', str(v.qry_start + 1), 'from file', self.final_assembly_read_depths, file=sys.stderr)
                            print(' - a variant was called at this position using nucmer, but there is no read depth (probably a mapping artifact)', file=sys.stderr)
                            depths = ['.'] * 4

                        ref_base, alt_base, ref_counts, alt_counts = depths

                        self.report_lines.append([
                            self.gene.id,
                            self.status_flag.to_number(),
                            total_reads,
                            self.name,
                            len(self.gene),
                            cov_per_contig[contig],
                            self.percent_identities[contig],
                            pymummer.variant.var_types[v.var_type],
                            effect,
                            new_bases,
                            v.ref_start + 1,
                            v.ref_end + 1,
                            v.ref_base,
                            v.qry_name,
                            v.qry_length,
                            v.qry_start + 1,
                            v.qry_end + 1,
                            v.qry_base,
                            ref_counts,
                            alt_base,
                            alt_counts,
                        ])

                        if contig in samtools_variants and v.qry_start in samtools_variants[contig]:
                            del samtools_variants[contig][v.qry_start]
                            if len(samtools_variants[contig]) == 0:
                                del samtools_variants[contig]

            if contig in samtools_variants:
                for pos in samtools_variants[contig]:
                    ref_base, alt_base, ref_counts, alt_counts = samtools_variants[contig][pos]
                    self.report_lines.append(
                      [
                        self.gene.id,
                        self.status_flag.to_number(),
                        total_reads,
                        self.name,
                        len(self.gene),
                        cov_per_contig[contig],
                        self.percent_identities[contig],
                      ] + \
                      ['.'] * 6 + \
                      [
                        contig,
                        len(self.final_assembly[contig]),
                        pos + 1,
                        pos + 1,
                        ref_base,
                        ref_counts,
                        alt_base,
                        alt_counts
                      ]
                    )

        if len(self.report_lines) == 0:
            for contig in self.percent_identities:
                self.report_lines.append([
                    self.gene.id,
                    self.status_flag.to_number(),
                    total_reads,
                    self.name,
                    len(self.gene),
                    cov_per_contig[contig],
                    self.percent_identities[contig],
                  ] + \
                  ['.'] * 6 + [contig, len(self.final_assembly[contig])] + ['.'] * 6
                )

        self.report_lines.sort(key=itemgetter(0, 14, 15))

    def _clean(self):
        if self.verbose:
            print('Cleaning', self.root_dir)

        if self.clean > 0:
            if self.verbose:
                print('  rm -r', self.assembly_dir)
            shutil.rmtree(self.assembly_dir)

        to_clean = [
            [
                'assembly.reads_mapped.unsorted.bam',
            ],
            [
                'assembly.fa.fai',
                'assembly.reads_mapped.bam.scaff',
                'assembly.reads_mapped.bam.soft_clipped',
                'assembly.reads_mapped.bam.unmapped_mates',
                'assembly_vs_gene.coords',
                'assembly_vs_gene.coords.snps',
                'genes.fa',
                'genes.fa.fai',
                'reads_1.fq',
                'reads_2.fq',
            ],
            [
                'assembly.fa.fai',
                'assembly.reads_mapped.bam',
                'assembly.reads_mapped.bam.vcf',
                'assembly_vs_gene.coords',
                'assembly_vs_gene.coords.snps',
            ]
        ]

        for i in range(self.clean + 1):
            for fname in to_clean[i]:
                fullname = os.path.join(self.root_dir, fname)
                if os.path.exists(fullname):
                    if self.verbose:
                        print('  rm', fname)
                    os.unlink(fullname)


    def run(self):
        self.gene = self._choose_best_gene()
        if self.gene is None:
            self.assembled_ok = False
        else:
            if self.assembler == 'velvet':
                self._assemble_with_velvet()
            elif self.assembler == 'spades':
                self._assemble_with_spades()

        # velvet can finish successfully, but make an empty contigs file
        if self.assembled_ok:
            number_of_contigs = pyfastaq.tasks.count_sequences(self.assembly_contigs)
            if number_of_contigs == 0:
                self.assembled_ok = False
                self.status_flag.add('assembly_fail')

        if self.assembled_ok:
            # finish the assembly
            self._scaffold_with_sspace()
            self._gap_fill_with_gapfiller()
            self._fix_contig_orientation()
            self._load_final_contigs()

            # map reads to assembly
            mapping.run_bowtie2(
                self.reads1,
                self.reads2,
                self.final_assembly_fa,
                self.final_assembly_bam[:-4],
                threads=self.threads,
                sort=True,
                samtools=self.samtools_exe,
                bowtie2=self.bowtie2_exe,
                bowtie2_preset=self.bowtie2_preset,
                verbose=self.verbose,
            )
            self._parse_assembly_bam()


            # compare gene and assembly
            self._run_nucmer(self.final_assembly_fa, self.assembly_vs_gene_coords, show_snps=True)
            self._parse_assembly_vs_gene_coords()
            self._nucmer_hits_to_percent_identity()
            self._get_mummer_variants()
            self._filter_mummer_variants()
            self._update_flag_from_nucmer_file()
            self._make_assembly_vcf()
            self._get_vcf_variant_counts()

        self._make_report_lines()
        self._clean()



import os
import sys
import shutil
import openpyxl
import pysam
import pyfastaq
from ariba import cdhit, cluster, common, mapping, histogram, faidx

class Error (Exception): pass


class Clusters:
    def __init__(self,
      db_fasta,
      reads_1,
      reads_2,
      outdir,
      assembly_kmer=21,
      threads=1,
      verbose=False,
      assembler='velvet',
      smalt_k=13,
      smalt_s=2,
      smalt_min_id=0.9,
      spades_other=None,
      max_insert=1000,
      min_scaff_depth=10,
      nucmer_min_id=90,
      nucmer_min_len=50,
      nucmer_breaklen=50,
      assembled_threshold=0.95,
      unique_threshold=0.03,
      bcftools_exe='bcftools',
      gapfiller_exe='GapFiller.pl',
      samtools_exe='samtools',
      smalt_exe='smalt',
      bowtie2_exe='bowtie2',
      bowtie2_preset='very-sensitive-local',
      spades_exe='spades.py',
      sspace_exe='SSPACE_Basic_v2.0.pl',
      velvet_exe='velvet', # prefix of velvet{g,h}
      cdhit_seq_identity_threshold=0.9,
      cdhit_length_diff_cutoff=0.9,
      run_cd_hit=True,
      clean=1,
    ):
        self.db_fasta = os.path.abspath(db_fasta)
        self.reads_1 = os.path.abspath(reads_1)
        self.reads_2 = os.path.abspath(reads_2)
        self.outdir = os.path.abspath(outdir)
        self.clusters_outdir = os.path.join(self.outdir, 'Clusters')
        self.clusters_info_file = os.path.join(self.outdir, 'clusters.tsv')
        self.clean = clean

        self.assembler = assembler
        assert self.assembler in ['velvet', 'spades']
        self.assembly_kmer = assembly_kmer
        self.spades_other = spades_other

        self.db_fasta_clustered = os.path.join(self.outdir, 'genes.clustered.fa')
        self.cluster_ids = {}
        self.bam_prefix = os.path.join(self.outdir, 'map_all_reads')
        self.bam = self.bam_prefix + '.bam'
        self.report_file_tsv = os.path.join(self.outdir, 'report.tsv')
        self.report_file_xls = os.path.join(self.outdir, 'report.xls')
        self.threads = threads
        self.verbose = verbose

        self.smalt_k = smalt_k
        self.smalt_s = smalt_s
        self.smalt_min_id = smalt_min_id
        self.max_insert = max_insert
        self.smalt_exe = smalt_exe
        self.bowtie2_exe = bowtie2_exe
        self.bowtie2_preset = bowtie2_preset

        self.insert_hist_bin = 10
        self.insert_hist = histogram.Histogram(self.insert_hist_bin)
        self.insert_size = None
        self.insert_sspace_sd = None
        self.insert_proper_pair_max = None

        self.min_scaff_depth = min_scaff_depth
        self.nucmer_min_id = nucmer_min_id
        self.nucmer_min_len = nucmer_min_len
        self.nucmer_breaklen = nucmer_breaklen

        self.assembled_threshold = assembled_threshold
        self.unique_threshold = unique_threshold

        self.cluster_to_dir = {}  # gene name -> abs path of cluster directory
        self.clusters = {}        # gene name -> Cluster object

        self.bcftools_exe = bcftools_exe

        self.sspace_exe = shutil.which(sspace_exe)
        if self.sspace_exe is None:
            print('WARNING: SSPACE not found. Scaffolding and gap filling will be skipped!', file=sys.stderr)
            self.gapfiller_exe = None
        else:
            self.sspace_exe = os.path.realpath(self.sspace_exe) # otherwise sspace dies loading packages
            self.gapfiller_exe = shutil.which(gapfiller_exe)
            if self.gapfiller_exe is None:
                print('WARNING: GapFiller not found. No gap filling will be run after scaffolding!', file=sys.stderr)
            else:
                self.gapfiller_exe = os.path.realpath(self.gapfiller_exe) # otherwise gapfiller dies loading packages

        self.samtools_exe = samtools_exe
        self.spades_exe = spades_exe

        self.velvet = velvet_exe

        self.cdhit_seq_identity_threshold = cdhit_seq_identity_threshold
        self.cdhit_length_diff_cutoff = cdhit_length_diff_cutoff
        self.run_cd_hit = run_cd_hit

        for d in [self.outdir, self.clusters_outdir]:
            try:
                os.mkdir(d)
            except:
                raise Error('Error mkdir ' + d)

    def _run_cdhit(self):
        r = cdhit.Runner(
            self.db_fasta,
            self.db_fasta_clustered,
            seq_identity_threshold=self.cdhit_seq_identity_threshold,
            threads=self.threads,
            length_diff_cutoff=self.cdhit_length_diff_cutoff,
            verbose=self.verbose,
        )
        if self.run_cd_hit:
            self.cluster_ids = r.run()
        else:
            if self.verbose:
                print('Skipping cd-hit because --no_cdhit option used')
            self.cluster_ids = r.fake_run()


    def _write_clusters_info_file(self):
        f = pyfastaq.utils.open_file_write(self.clusters_info_file)
        print('#Cluster\tGene', file=f)
        for c in sorted([int(x) for x in self.cluster_ids]):
            for seqname in sorted(list(self.cluster_ids[str(c)])):
                print(c, seqname, sep='\t', file=f)
        pyfastaq.utils.close(f)


    def _map_reads_to_clustered_genes(self):
        mapping.run_bowtie2(
            self.reads_1,
            self.reads_2,
            self.db_fasta_clustered,
            self.bam_prefix,
            threads=self.threads,
            samtools=self.samtools_exe,
            bowtie2=self.bowtie2_exe,
            bowtie2_preset=self.bowtie2_preset,
            verbose=self.verbose,
        )


    def _sam_to_fastq(self, s):
        name = s.qname
        if s.is_read1:
            name += '/1'
        elif s.is_read2:
            name += '/2'
        else:
            raise Error('Read ' + name + ' must be first or second of pair according to flag. Cannot continue')

        seq = pyfastaq.sequences.Fastq(name, common.decode(s.seq), common.decode(s.qual))
        if s.is_reverse:
            seq.revcomp()

        return seq


    def _sam_pair_to_insert(self, s1, s2):
        if s1.is_unmapped or s2.is_unmapped or (s1.tid != s2.tid) or (s1.is_reverse == s2.is_reverse):
            return None

        # If here, reads are both mapped to the same ref, and in opposite orientations
        if s1.is_reverse:
            end = s1.reference_end - 1
            start = s2.reference_start
        else:
            end = s2.reference_end - 1
            start = s1.reference_start

        if start < end:
            return end - start + 1
        else:
            return None


    def _bam_to_clusters_reads(self):
        '''Sets up Cluster directories (one for each gene that has reads that mapped to it), writes reads fwd and rev files. Also gathers histogram data of insert size'''
        filehandles_1 = {} # gene name -> filehandle of fwd_reads
        filehandles_2 = {} # gene name -> filehandle of rev_reads
        sam_reader = pysam.Samfile(self.bam, "rb")
        sam1 = None

        for s in sam_reader.fetch(until_eof=True):
            if sam1 is None:
                sam1 = s
                continue

            ref_seqs = set()
            if not s.is_unmapped:
                ref_seqs.add(sam_reader.getrname(s.tid))
            if not sam1.is_unmapped:
                ref_seqs.add(sam_reader.getrname(sam1.tid))

            read1 = self._sam_to_fastq(sam1)
            read2 = self._sam_to_fastq(s)
            if read1.id.endswith('/2'):
                read1, read2 = read2, read1

            insert = self._sam_pair_to_insert(s, sam1)
            if insert is not None:
                self.insert_hist.add(insert)

            for ref in ref_seqs:
                if ref not in self.cluster_to_dir:
                    assert ref not in filehandles_1
                    assert ref not in filehandles_2

                    new_dir = os.path.join(self.clusters_outdir, ref)
                    try:
                        os.mkdir(new_dir)
                    except:
                        raise Error('Error mkdir ' + new_dir)

                    self.cluster_to_dir[ref] = new_dir
                    filehandles_1[ref] = pyfastaq.utils.open_file_write(os.path.join(new_dir, 'reads_1.fq'))
                    filehandles_2[ref] = pyfastaq.utils.open_file_write(os.path.join(new_dir, 'reads_2.fq'))
                    if self.verbose:
                        print('New cluster with reads that hit:', ref, flush=True)

                print(read1, file=filehandles_1[ref])
                print(read2, file=filehandles_2[ref])

            sam1 = None

        for ref in filehandles_1:
            pyfastaq.utils.close(filehandles_1[ref])
            pyfastaq.utils.close(filehandles_2[ref])

        if self.verbose:
            print('Total clusters to perform local assemblies:', len(self.cluster_to_dir), flush=True)

    def _set_insert_size_data(self):
        assert len(self.insert_hist) > 0
        (x, self.insert_size, pc95, self.insert_sspace_sd) = self.insert_hist.stats()
        self.insert_sspace_sd = min(1, self.insert_sspace_sd)
        self.insert_proper_pair_max = 1.1 * pc95
        if self.verbose:
            print('\nInsert size information from reads mapped to reference genes:')
            print('Insert size:', self.insert_size, sep='\t')
            print('Insert sspace sd:', self.insert_sspace_sd, sep='\t')
            print('Max insert:', self.insert_proper_pair_max, sep='\t')
            print()


    def _init_and_run_clusters(self):
        if len(self.cluster_to_dir) == 0:
            raise Error('Did not get any reads mapped to genes. Cannot continue')

        counter = 0

        for gene in sorted(self.cluster_to_dir):
            counter += 1
            if self.verbose:
                print('\nAssembling cluster', counter, 'of', str(len(self.cluster_to_dir)))
            new_dir = self.cluster_to_dir[gene]

            faidx.write_fa_subset(
                self.cluster_ids[gene],
                self.db_fasta,
                os.path.join(new_dir, 'genes.fa'),
                samtools_exe=self.samtools_exe,
                verbose=self.verbose
            )

            self.clusters[gene] = cluster.Cluster(
                new_dir,
                gene,
                assembly_kmer=self.assembly_kmer,
                assembler=self.assembler,
                max_insert=self.insert_proper_pair_max,
                min_scaff_depth=self.min_scaff_depth,
                nucmer_min_id=self.nucmer_min_id,
                nucmer_min_len=self.nucmer_min_len,
                nucmer_breaklen=self.nucmer_breaklen,
                sspace_k=self.min_scaff_depth,
                reads_insert=self.insert_size,
                sspace_sd=self.insert_sspace_sd,
                threads=self.threads,
                assembled_threshold=self.assembled_threshold,
                unique_threshold=self.unique_threshold,
                verbose=self.verbose,
                bcftools_exe=self.bcftools_exe,
                gapfiller_exe=self.gapfiller_exe,
                samtools_exe=self.samtools_exe,
                bowtie2_exe=self.bowtie2_exe,
                bowtie2_preset=self.bowtie2_preset,
                spades_exe=self.spades_exe,
                sspace_exe=self.sspace_exe,
                velvet_exe=self.velvet,
                spades_other=self.spades_other,
                clean=self.clean,
            )

            self.clusters[gene].run()


    def _write_reports(self):
        columns = [
            '#gene',
            'flag',
            'reads',
            'cluster',
            'gene_len',
            'assembled',
            'pc_ident',
            'var_type',
            'var_effect',
            'new_aa',
            'gene_start',
            'gene_end',
            'gene_nt',
            'scaffold',
            'scaff_len',
            'scaff_start',
            'scaff_end',
            'scaff_nt',
            'read_depth',
            'alt_bases',
            'ref_alt_depth'
        ]

        f = pyfastaq.utils.open_file_write(self.report_file_tsv)
        print('\t'.join(columns), file=f)

        columns[0] = 'gene'
        workbook = openpyxl.Workbook()
        worksheet = workbook.worksheets[0]
        worksheet.title = 'ARIBA_report'
        worksheet.append(columns)

        for gene in sorted(self.clusters):
            for line in self.clusters[gene].report_lines:
                print('\t'.join([str(x) for x in line]), file=f)
                worksheet.append(line)
        pyfastaq.utils.close(f)
        workbook.save(self.report_file_xls)


    def _clean(self):
        to_clean = [
            [
            ],
            [
                self.bam
            ],
            [
                self.db_fasta_clustered,
                self.db_fasta_clustered + '.fai',
                self.clusters_info_file,
            ]
        ]

        for i in range(self.clean + 1):
            for fname in to_clean[i]:
                if os.path.exists(fname):
                    if self.verbose:
                        print('  rm', fname)
                    os.unlink(fname)

        if self.clean >= 2:
            if self.verbose:
                print('  rm -r', self.clusters_outdir)
                shutil.rmtree(self.clusters_outdir)


    def run(self):
        cwd = os.getcwd()
        os.chdir(self.outdir)

        if self.verbose:
            print('{:_^79}'.format(' Running cd-hit '), flush=True)
        self._run_cdhit()
        self._write_clusters_info_file()
        if self.verbose:
            print('Finished cd-hit\n')
            print('{:_^79}'.format(' Mapping reads to clustered genes '), flush=True)
        self._map_reads_to_clustered_genes()
        if self.verbose:
            print('Finished mapping\n')
            print('{:_^79}'.format(' Generating clusters '), flush=True)
        self._bam_to_clusters_reads()
        self._set_insert_size_data()
        if self.verbose:
            print('{:_^79}'.format(' Assembling each cluster '), flush=True)
        self._init_and_run_clusters()
        if self.verbose:
            print('Finished assembling clusters\n')
            print('{:_^79}'.format(' Writing report files '), flush=True)
        self._write_reports()
        if self.verbose:
            print('Finished writing report files. Cleaning files', flush=True)
        self._clean()
        if self.verbose:
            print('\nAll done!\n')

        os.chdir(cwd)



import os
import pysam
from ariba import common

class Error (Exception): pass


def run_bowtie2(
      reads_fwd,
      reads_rev,
      ref_fa,
      out_prefix,
      threads=1,
      max_insert=1000,
      sort=False,
      samtools='samtools',
      bowtie2='bowtie2',
      bowtie2_preset='very-sensitive-local',
      verbose=False
    ):

    map_index = out_prefix + '.map_index'
    clean_files = [map_index + '.' + x + '.bt2' for x in ['1', '2', '3', '4', 'rev.1', 'rev.2']]
    index_cmd = ' '.join([
        bowtie2 + '-build',
        '-q',
        ref_fa,
        map_index
    ])

    final_bam = out_prefix + '.bam'
    if sort:
        intermediate_bam = out_prefix + '.unsorted.bam'
    else:
        intermediate_bam = final_bam

    map_cmd = ' '.join([
        bowtie2,
        '--threads', str(threads),
        '--' + bowtie2_preset,
        '-X', str(max_insert),
        '-x', map_index,
        '-1', reads_fwd,
        '-2', reads_rev,
        '|', samtools, 'view',
        '-bS -T', ref_fa,
        '- >', intermediate_bam
    ])

    common.syscall(index_cmd, verbose=verbose)
    common.syscall(map_cmd, verbose=verbose)

    if sort:
        threads = min(4, threads)
        thread_mem = int(500 / threads)
        sort_cmd = samtools + ' sort -@' + str(threads) + ' -m ' + str(thread_mem) + 'M ' + intermediate_bam + ' ' + out_prefix
        index_cmd = samtools + ' index ' + final_bam
        common.syscall(sort_cmd, verbose=verbose)
        common.syscall(index_cmd, verbose=verbose)
    for fname in clean_files:
        os.unlink(fname)


def run_smalt(
      reads_fwd,
      reads_rev,
      ref_fa,
      out_prefix,
      index_k=9,
      index_s=2,
      threads=1,
      max_insert=1000,
      minid=0.9,
      sort=False,
      extra_smalt_map_ops='-x',
      samtools='samtools',
      smalt='smalt',
      verbose=False
    ):
    if extra_smalt_map_ops is None:
        extra_smalt_map_ops = ''
    map_index = out_prefix + '.map_index'
    clean_files = [map_index + '.' + x for x in ['smi', 'sma']]
    index_cmd = ' '.join([
        smalt, 'index',
        '-k', str(index_k),
        '-s', str(index_s),
        map_index,
        ref_fa
    ])

    map_cmd = smalt + ' map ' + extra_smalt_map_ops + ' '

    # depending on OS, -n can break smalt, so only use -n if it's > 1.
    if threads > 1:
        map_cmd += '-n ' + str(threads) + ' -O '

    if reads_rev is None:
        map_cmd += ' '.join([
            '-y', str(minid),
            map_index,
            reads_fwd,
        ])
    else:
        map_cmd += ' '.join([
            '-i', str(max_insert),
            '-y', str(minid),
            map_index,
            reads_fwd,
            reads_rev,
        ])

    map_cmd += ' | ' + samtools + ' view'

    final_bam = out_prefix + '.bam'
    if sort:
        intermediate_bam = out_prefix + '.unsorted.bam'
    else:
        intermediate_bam = final_bam

    map_cmd += ' -bS -T ' + ref_fa + '  - > ' + intermediate_bam
    common.syscall(index_cmd, verbose=verbose)
    common.syscall(map_cmd, verbose=verbose)

    if sort:
        threads = min(4, threads)
        thread_mem = int(500 / threads)
        sort_cmd = samtools + ' sort -@' + str(threads) + ' -m ' + str(thread_mem) + 'M ' + intermediate_bam + ' ' + out_prefix
        index_cmd = samtools + ' index ' + final_bam
        common.syscall(sort_cmd, verbose=verbose)
        common.syscall(index_cmd, verbose=verbose)
    for fname in clean_files:
        os.unlink(fname)


def get_total_alignment_score(bam):
    '''Returns total of AS: tags in the input BAM'''
    sam_reader = pysam.Samfile(bam, "rb")
    total = 0
    for sam in sam_reader.fetch(until_eof=True):
        try:
            total += sam.opt('AS')
        except:
            pass
    return total




import os
from ariba import common


def write_fa_subset(seq_names, infile, outfile, samtools_exe='samtools', verbose=False):
    if not os.path.exists(infile + '.fai'):
        common.syscall(samtools_exe + ' faidx ' + infile, verbose=verbose)

    if os.path.exists(outfile):
        os.path.unlink(outfile)

    for name in seq_names:
        common.syscall(' '.join([
            samtools_exe + ' faidx',
            infile,
            '"' + name + '"',
            '>>', outfile
        ]))




import argparse
import pyfastaq
import ariba


def run():
    parser = argparse.ArgumentParser(
        description = 'ARIBA: Antibiotic Resistance Identification By Assembly',
        usage = 'ariba run [options] <db.fa> <reads1.fq> <reads2.fq> <outdir>')
    parser.add_argument('db_fasta', help='FASTA file of reference genes')
    parser.add_argument('reads_1', help='Name of fwd reads fastq file')
    parser.add_argument('reads_2', help='Name of rev reads fastq file')
    parser.add_argument('outdir', help='Output directory (must not already exist)')

    cdhit_group = parser.add_argument_group('cd-hit options')
    cdhit_group.add_argument('--no_cdhit', action='store_true', help='Do not run cd-hit')
    cdhit_group.add_argument('--cdhit_seq_identity_threshold', type=float, help='Sequence identity threshold (cd-hit option -c) [%(default)s]', default=0.9, metavar='FLOAT')
    cdhit_group.add_argument('--cdhit_length_diff_cutoff', type=float, help='length difference cutoff (cd-hit option -s) [%(default)s]', default=0.9, metavar='FLOAT')

    nucmer_group = parser.add_argument_group('nucmer options')
    nucmer_group.add_argument('--nucmer_min_id', type=int, help='Minimum alignment identity (delta-filter -i) [%(default)s]', default=90, metavar='INT')
    nucmer_group.add_argument('--nucmer_min_len', type=int, help='Minimum alignment length (delta-filter -i) [%(default)s]', default=50, metavar='INT')
    nucmer_group.add_argument('--nucmer_breaklen', type=int, help='Value to use for -breaklen when running nucmer [%(default)s]', default=50, metavar='INT')

    assembly_group = parser.add_argument_group('Assembly options')
    allowed_assemblers = ['velvet', 'spades']
    assembly_group.add_argument('--assembler', help='Assembler to use. Available options: ' + ','.join(allowed_assemblers) + ' [%(default)s]', choices=allowed_assemblers, default='spades', metavar='Assembler')
    assembly_group.add_argument('--min_scaff_depth', type=int, help='Minimum number of read pairs needed as evidence for scaffold link between two contigs. This is also the value used for sspace -k when scaffolding [%(default)s]', default=10, metavar='INT')
    assembly_group.add_argument('--assembler_k', type=int, help='kmer size to use with assembler. You can use 0 to set kmer to 2/3 of the read length. Warning - lower kmers are usually better. [%(default)s]', metavar='INT', default=21)
    assembly_group.add_argument('--spades_other', help='Put options string to be used with spades in quotes. This will NOT be sanity checked. Do not use -k or -t: for these options you should use the ariba run options --assembler_k and --threads [%(default)s]', default="--only-assembler", metavar="OPTIONS")

    other_group = parser.add_argument_group('Other options')
    other_group.add_argument('--genetic_code', type=int, help='Number of genetic code to use. Currently supported 1,4,11 [%(default)s]', choices=[1,4,11], default=11, metavar='INT')
    other_group.add_argument('--threads', type=int, help='Number of threads for bowtie2 and spades [%(default)s]', default=1, metavar='INT')
    bowtie2_presets = ['very-fast-local', 'fast-local', 'sensitive-local', 'very-sensitive-local']
    other_group.add_argument('--bowtie2_preset', choices=bowtie2_presets, help='Preset option for bowtie2 mapping [%(default)s]', default='very-sensitive-local', metavar='|'.join(bowtie2_presets))
    other_group.add_argument('--assembled_threshold', type=float, help='If proportion of gene assembled (regardless of into how many contigs) is at least this value then the flag gene_assembled is set [%(default)s]', default=0.95, metavar='FLOAT (between 0 and 1)')
    other_group.add_argument('--unique_threshold', type=float, help='If proportion of bases in gene assembled more than once is <= this value, then the flag unique_contig is set [%(default)s]', default=0.03, metavar='FLOAT (between 0 and 1)')
    other_group.add_argument('--clean', type=int, choices=[0,1,2], help='Specify how much cleaning to do. 0=none, 1=some, 2=only keep the report [%(default)s]', default=1, metavar='INT')
    other_group.add_argument('--verbose', action='store_true', help='Be verbose')

    executables_group = parser.add_argument_group('executables locations')
    executables_group.add_argument('--bcftools', help='bcftools executable [bcftools]', metavar='PATH')
    executables_group.add_argument('--bowtie2', help='bowtie2 executable [bowtie2]', metavar='PATH')
    executables_group.add_argument('--cdhit', help=argparse.SUPPRESS)
    executables_group.add_argument('--gapfiller', help='GapFiller executable [GapFiller.pl]', metavar='PATH')
    executables_group.add_argument('--nucmer', help=argparse.SUPPRESS, default='nucmer')
    executables_group.add_argument('--samtools', help='samtools executable [samtools]', metavar='PATH')
    executables_group.add_argument('--spades', help='SPAdes executable [spades.py]',  metavar='PATH')
    executables_group.add_argument('--sspace', help='SSPACE executable [SSPACE_Basic_v2.0.pl]', metavar='PATH')
    executables_group.add_argument('--velvet', help='prefix of velvet{g,h} executables [velvet]', metavar='PATH')
    executables_group.add_argument('--velvetg', help=argparse.SUPPRESS)
    executables_group.add_argument('--velveth', help=argparse.SUPPRESS)

    options = parser.parse_args()
    if options.assembler == 'velvet':
        options.velvet = 'velvet'
    ariba.external_progs.check_versions(options, verbose=options.verbose, not_required=set(['sspace', 'gapfiller']))
    pyfastaq.sequences.genetic_code = options.genetic_code

    c = ariba.clusters.Clusters(
          options.db_fasta,
          options.reads_1,
          options.reads_2,
          options.outdir,
          assembly_kmer=options.assembler_k,
          assembler=options.assembler,
          threads=options.threads,
          verbose=options.verbose,
          min_scaff_depth=options.min_scaff_depth,
          nucmer_min_id=options.nucmer_min_id,
          nucmer_min_len=options.nucmer_min_len,
          nucmer_breaklen=options.nucmer_breaklen,
          spades_other=options.spades_other,
          assembled_threshold=options.assembled_threshold,
          unique_threshold=options.unique_threshold,
          bcftools_exe=options.bcftools,
          gapfiller_exe=options.gapfiller,
          samtools_exe=options.samtools,
          bowtie2_exe=options.bowtie2,
          bowtie2_preset=options.bowtie2_preset,
          spades_exe=options.spades,
          sspace_exe=options.sspace,
          velvet_exe=options.velvet,
          cdhit_seq_identity_threshold=options.cdhit_seq_identity_threshold,
          cdhit_length_diff_cutoff=options.cdhit_length_diff_cutoff,
          clean=options.clean,
          run_cd_hit=(not options.no_cdhit)
        )
    c.run()




import argparse
import ariba

def run():
    parser = argparse.ArgumentParser(
        description = 'Translate the meaning of a flag output by ARIBA',
        usage = 'ariba flag <flag>')
    parser.add_argument('flag_in', type=int, help='Flag to be translated (an integer)', metavar='flag')
    options = parser.parse_args()

    f = ariba.flag.Flag(options.flag_in)
    print('Meaning of flag', f)
    print(f.to_long_string())



class Error (Exception): pass

class Histogram:
    def __init__(self, bin_width=10):
        self.bin_width = bin_width
        self.bins = {}


    def __len__(self):
        if len(self.bins) == 0:
            return 0
            
        return sum(list(self.bins.values()))

    def _to_bin(self, n):
        return self.bin_width * (n // self.bin_width)


    def add(self, n):
        b = self._to_bin(n)
        self.bins[b] = self.bins.get(b, 0) + 1


    def stats(self):
        if len(self.bins) == 0:
            return None

        total = sum([self.bins[x] for x in self.bins])

        pc5 = None
        pc95 = None
        median = None
        sspace_sd = None
        i = 0

        for x in sorted(self.bins):
            i += self.bins[x]
            if i >= 0.05 * total and pc5 is None:
                pc5 = x + 0.5 * self.bin_width
            if i >= 0.5 * total and median is None:
                median = x + 0.5 * self.bin_width
            if i >= 0.95 * total and pc95 is None:
                pc95 = x + 0.5 + self.bin_width

        if None not in [pc5, pc95, median]:
            sspace_sd = round(max(pc95 - median, median - pc5) / median, 2)
 
        return (pc5, median, pc95, sspace_sd)




import os
import openpyxl
import pyfastaq
from ariba import flag

class Error (Exception): pass

columns = [
    'gene',
    'flag',
    'reads',
    'cluster',
    'gene_len',
    'assembled',
    'pc_ident',
    'var_type',
    'var_effect',
    'new_aa',
    'gene_start',
    'gene_end',
    'gene_nt',
    'scaffold',
    'scaff_len',
    'scaff_start',
    'scaff_end',
    'scaff_nt',
    'read_depth',
    'alt_bases',
    'ref_alt_depth'
]

int_columns = [
    'reads',
    'gene_len',
    'assembled',
    'gene_start',
    'gene_end',
    'scaff_len',
    'scaff_start',
    'scaff_end',
    'read_depth',
]


class Summary:
    def __init__(
      self,
      outfile,
      filenames=None,
      fofn=None,
      min_id=90.0
    ):
        if filenames is None and fofn is None:
            raise Error('Error! Must supply filenames or fofn to Summary(). Cannot continue')

        if filenames is None:
            self.filenames = []
        else:
            self.filenames = filenames

        if fofn is not None:
            self.filenames.extend(self._load_fofn(fofn))

        self.min_id = min_id
        self.outfile = outfile


    def _load_fofn(self, fofn):
        f = pyfastaq.utils.open_file_read(fofn)
        filenames = [x.rstrip() for x in f.readlines()]
        pyfastaq.utils.close(f)
        return filenames


    def _check_files_exist(self):
        for fname in self.filenames:
            if not os.path.exists(fname):
                raise Error('File not found: "' + fname + '". Cannot continue')


    def _line2dict(self, line):
        data = line.rstrip().split('\t')
        d = {columns[i]: data[i] for i in range(len(data))}
        d['flag'] = flag.Flag(int(d['flag']) )
        for key in int_columns:
            try:
                d[key] = int(d[key])
            except:
                assert d[key] == '.'
        try:
            d['pc_ident'] = float(d['pc_ident'])
        except:
            assert d['pc_ident'] == '.'
        return d


    def _load_file(self, filename):
        f = pyfastaq.utils.open_file_read(filename)
        d = {}

        for line in f:
            if line.startswith('#'):
                if line.rstrip()[1:].split('\t') != columns:
                    raise Error('Error parsing the following line.\n' + line)
                continue
            data = self._line2dict(line)

            if data['gene'] not in d:
                d[data['gene']] = []

            d[data['gene']].append(data)

        pyfastaq.utils.close(f)
        return d


    def _to_summary_number(self, l):
        f = l[0]['flag']
        if f.has('assembly_fail') or not f.has('gene_assembled') or self._pc_id_of_longest(l) <= self.min_id:
            return 0

        if f.has('hit_both_strands') or (not f.has('complete_orf')):
            return 1

        if f.has('unique_contig') and f.has('gene_assembled_into_one_contig'):
            return 3

        return 2


    def _pc_id_of_longest(self, l):
        longest = 0
        identity = None
        for data in l:
            if data['assembled'] > longest:
                longest = data['assembled']
                identity = data['pc_ident']

        assert identity is not None
        return identity



    def _gather_output_rows(self):
        self.data = {filename: self._load_file(filename) for filename in self.filenames}

        all_genes = set()
        for l in self.data.values():
            all_genes.update(set(l.keys()))
        all_genes = list(all_genes)
        all_genes.sort()

        self.rows_out = []
        self.rows_out.append(['filename'] + all_genes)

        for filename in self.filenames:
            new_row = [filename]
            for gene in all_genes:
                if gene not in self.data[filename]:
                    new_row.append(0)
                else:
                    new_row.append(self._to_summary_number(self.data[filename][gene]))

            self.rows_out.append(new_row)


    def _filter_output_rows(self):
        # remove rows that are all zeros
        self.rows_out = [x for x in self.rows_out if x[1:] != [0]*(len(x)-1)]

        # remove columns that are all zeros
        to_remove = []
        for i in range(1, len(self.rows_out[0])):
            if sum([x[i] for x in self.rows_out[1:]]) == 0:
                to_remove.append(i)

        for i in range(len(self.rows_out)):
            self.rows_out[i] = [self.rows_out[i][j] for j in range(len(self.rows_out[i])) if j not in to_remove]



    def _write_tsv(self):
        f = pyfastaq.utils.open_file_write(self.outfile)
        print('#', end='', file=f)
        for row in self.rows_out:
            print('\t'.join([str(x) for x in row]), file=f)
        pyfastaq.utils.close(f)


    def _write_xls(self):
        workbook = openpyxl.Workbook()
        worksheet = workbook.worksheets[0]
        worksheet.title = 'ARIBA_summary'
        for row in self.rows_out:
            worksheet.append(row)
        workbook.save(self.outfile)


    def run(self):
        self._check_files_exist()
        self._gather_output_rows()
        self._filter_output_rows()
        if self.outfile.endswith('.xls'):
            self._write_xls()
        else:
            self._write_tsv()





import tempfile
import shutil
import os
import pyfastaq
from ariba import common

class Error (Exception): pass



class Runner:
    def __init__(
      self,
      infile,
      outfile,
      seq_identity_threshold=0.9,
      threads=1,
      length_diff_cutoff=0.9,
      verbose=False,
    ):

        if not os.path.exists(infile):
            raise Error('File not found: "' + infile + '". Cannot continue')

        self.infile = os.path.abspath(infile)
        self.outfile = os.path.abspath(outfile)
        self.seq_identity_threshold = seq_identity_threshold
        self.threads = threads
        self.length_diff_cutoff = length_diff_cutoff
        self.verbose = verbose


    def fake_run(self):
        '''Doesn't actually run cd-hit. Instead, puts each input sequence into its own cluster. So it's as if cdhit was run, but didn't cluster anything'''
        cluster_to_name = {}
        found_names = set()
        seq_reader = pyfastaq.sequences.file_reader(self.infile)
        f = pyfastaq.utils.open_file_write(self.outfile)
        for seq in seq_reader:
            if seq.id in found_names:
                raise Error('Sequence name "' + seq.id + '" not unique. Cannot continue')
            found_names.add(seq.id)
            cluster_number = str(len(cluster_to_name))
            cluster_to_name[cluster_number] = {seq.id}
            seq.id = cluster_number
            print(seq, file=f)

        pyfastaq.utils.close(f)
        return cluster_to_name


    def run(self):
        tmpdir = tempfile.mkdtemp(prefix='tmp.run_cd-hit.', dir=os.getcwd())
        cdhit_fasta = os.path.join(tmpdir, 'cdhit')
        cluster_info_outfile = cdhit_fasta + '.bak.clstr'
        infile_renamed = os.path.join(tmpdir, 'input.renamed.fa')

        # cd-hit truncates all names to 19 bases in its report of which
        # sequences belong to which clusters. So need to temporarily
        # rename all sequences to have short enough names. Grrr.
        new_to_old_name = self._enumerate_fasta(self.infile, infile_renamed)

        cmd = ' '.join([
            'cd-hit-est',
            '-i', infile_renamed,
            '-o', cdhit_fasta,
            '-c', str(self.seq_identity_threshold),
            '-T', str(self.threads),
            '-s', str(self.length_diff_cutoff),
            '-bak 1',
        ])

        common.syscall(cmd, verbose=self.verbose)

        cluster_representatives = self._get_ids(cdhit_fasta)
        clusters, cluster_rep_to_cluster = self._parse_cluster_info_file(cluster_info_outfile, new_to_old_name, cluster_representatives)
        self._rename_fasta(cdhit_fasta, self.outfile, cluster_rep_to_cluster)
        shutil.rmtree(tmpdir)
        return clusters


    def _enumerate_fasta(self, infile, outfile):
        rename_file = outfile + '.tmp.rename_info'
        assert not os.path.exists(rename_file)
        pyfastaq.tasks.enumerate_names(infile, outfile, rename_file=rename_file)

        with open(rename_file) as f:
            lines = [x.rstrip().split('\t') for x in f.readlines() if x != '#old\tnew\n']
            new_to_old_name = {x[1]: x[0] for x in lines}
            if len(lines) != len(new_to_old_name):
                raise Error('Sequence names in input file not unique! Cannot continue')

        os.unlink(rename_file)
        return new_to_old_name


    def _rename_fasta(self, infile, outfile, names_dict):
        seq_reader = pyfastaq.sequences.file_reader(infile)
        f = pyfastaq.utils.open_file_write(outfile)
        for seq in seq_reader:
            seq.id = names_dict[seq.id]
            print(seq, file=f)

        pyfastaq.utils.close(f)


    def _parse_cluster_info_file(self, infile, names_dict, cluster_representatives):
        f = pyfastaq.utils.open_file_read(infile)
        clusters = {}
        cluster_representative_to_cluster_number = {}
        for line in f:
            data = line.rstrip().split()
            cluster = data[0]
            seqname = data[2]
            if not (seqname.startswith('>') and seqname.endswith('...')):
                raise Error('Unexpected format of sequence name in line:\n' + line)
            seqname = seqname[1:-3]

            if seqname in cluster_representatives:
                cluster_representative_to_cluster_number[seqname] = cluster

            seqname = names_dict[seqname]

            if cluster not in clusters:
                clusters[cluster] = set()

            if seqname in clusters[cluster]:
                raise Error('Duplicate name "' + seqname + '" found in cluster ' + str(cluster))

            clusters[cluster].add(seqname)

        pyfastaq.utils.close(f)

        return clusters, cluster_representative_to_cluster_number


    def _get_ids(self, infile):
        seq_reader = pyfastaq.sequences.file_reader(infile)
        return set([seq.id for seq in seq_reader])




import shutil
import subprocess
import os
from distutils.version import LooseVersion
import re
import sys
import pyfastaq
from ariba import common

class Error (Exception): pass

def is_in_path(prog):
    return shutil.which(prog) is not None


prog_to_default = {
    'bcftools': 'bcftools',
    'bowtie2': 'bowtie2',
    'cdhit': 'cd-hit-est',
    'gapfiller': 'GapFiller.pl',
    'nucmer' : 'nucmer',
    'samtools': 'samtools',
    'smalt': 'smalt',
    'spades': 'spades.py',
    'sspace': 'SSPACE_Basic_v2.0.pl',
    'velvetg': 'velvetg',
    'velveth': 'velveth',
}


prog_to_env_var = {
    'bcftools': 'ARIBA_BCFTOOLS',
    'samtools': 'ARIBA_SAMTOOLS',
    'spades': 'ARIBA_SPADES', 
}


prog_to_version_cmd = {
    'bcftools': ('', re.compile('^Version: ([0-9\.]+)')),
    'bowtie2': ('--version', re.compile('.*bowtie2-align version (.*)$')),
    'cdhit': ('', re.compile('CD-HIT version ([0-9\.]+) \(')),
    'gapfiller': ('', re.compile('^Usage: .*pl \[GapFiller_(.*)\]')),
    'nucmer': ('--version', re.compile('^NUCmer \(NUCleotide MUMmer\) version ([0-9\.]+)')),
    'samtools': ('', re.compile('^Version: ([0-9\.]+)')),
    'smalt': ('version', re.compile('^Version: ([0-9\.]+)')),
    'spades': ('', re.compile('^SPAdes genome assembler v.([0-9\.]+)')),
    'sspace': ('', re.compile('^Usage: .*pl \[SSPACE_(.*)\]')),
    'velvetg': ('', re.compile('Version ([0-9\.]+)')),
    'velveth': ('', re.compile('Version ([0-9\.]+)')),
}


min_versions = {
    'bcftools': '1.2',
    'bowtie2': '2.1.0',
    'cd-hit': '4.6',
    'nucmer': '3.1',
    'samtools': '1.2',
    'smalt': '0.7.4',
    'spades': '3.5.0',
    'velvetg': '1.2.07',
    'velveth': '1.2.07',
}


def set_path(prog, opts):
    path_from_opts = eval('opts.' + prog)
    if path_from_opts is not None:
        return

    if prog in prog_to_env_var:
        env_var = prog_to_env_var[prog]
        if env_var in os.environ:
            exec('opts.' + prog + ' = "' + os.environ[env_var] + '"')
            return

    exec('opts.' + prog + ' = "' + prog_to_default[prog] + '"')


def get_version(prog, path=None, raise_error=True):
    assert prog in prog_to_version_cmd
    if path is None:
        path = prog

    if not is_in_path(path):
        if raise_error:
            raise Error('Error getting version of ' + path + ' - not found in path.')
        else:
            return 'Not_in_path', 'Not_in_path'

    path = shutil.which(path)

    if prog in ['sspace', 'gapfiller']:
        cmd = 'perl ' + os.path.realpath(shutil.which(path))
        regex = prog_to_version_cmd[prog][1]
    else:
        cmd, regex = prog_to_version_cmd[prog]
        cmd = path + ' ' + cmd

    cmd_output = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()
    cmd_output = common.decode(cmd_output[0]).split('\n')[:-1] + common.decode(cmd_output[1]).split('\n')[:-1]
    for line in cmd_output:
        hits = regex.search(line)
        if hits:
            return hits.group(1), path
    return 'UNKNOWN ...\n I tried running this to get the version: "' + cmd + '"\n and the output didn\'t match this regular expression: "' + regex.pattern + '"', path


def check_versions(opts, verbose=False, not_required=None):
    if not_required is None:
        not_required = set()

    if verbose:
        print('{:_^79}'.format(' Checking dependencies and their versions '))
        print('tool', 'version', 'path', sep='\t')

    to_check = [
        'bcftools',
        'bowtie2',
        'cdhit',
        'nucmer',
        'samtools',
        'sspace',
        'gapfiller',
    ]
    
    if opts.assembler == 'spades':
        to_check.append('spades')
    elif opts.assembler == 'velvet':
        to_check.append('velvetg')
        to_check.append('velveth')
    else:
        raise Error('Assembler ' + opts.assembler + ' not recognised. Cannot continue')

    errors = []
    failed_to_find = set()

    for prog in to_check:
        set_path(prog, opts)
        version, path = get_version(prog, path=eval('opts.' + prog), raise_error=prog not in not_required)
        if verbose:
            print(prog, version, path, sep='\t')
        if path == 'Not_in_path':
            print('\nWARNING:', prog, 'not found in path, so will be skipped during assembly\n', file=sys.stderr)

        if prog in min_versions and LooseVersion(version) < LooseVersion(min_versions[prog]):
            errors.append(' '.join(['Found version', version, 'of', prog, 'which is too low! Please update to at least', min_versions[prog] + '\n   Found it here:', path]))
            failed_to_find.add(prog)

    if len(errors):
        for e in errors:
            print('\n*** Error! Bad dependency! ***', file=sys.stderr)
            print(e, file=sys.stderr)
            print()
        if len(failed_to_find.difference(not_required)) > 0:
            raise Error('Cannot continue. Some dependencies need updating')
        else:
            assert failed_to_find.issubset(not_required)
            if 'sspace' in failed_to_find:
                print('WARNING: SSPACE not found. Will not run scaffolding or gap filling', file=sys.stderr)
            elif 'gapfiller' in failed_to_find:
                print('WARNING: GapFiller not found. Will not run gap filling after scaffolding', file=sys.stderr)

    if verbose:
        print('\nDependencies look OK (but check in case there are warnings about SSPACE or GapFiller)\n')



import argparse
import sys
import pyfastaq
import ariba

def run():
    parser = argparse.ArgumentParser(
        description = 'Check or fix resistance genes FASTA file',
        usage = 'ariba refcheck [options] <infile>')
    parser.add_argument('--genetic_code', type=int, help='Number of genetic code to use. Currently supported 1,4,11 [%(default)s]', choices=[1,4,11], default=11, metavar='INT')
    parser.add_argument('-m', '--min_length', type=int, help='Minimum length in nucleotides of gene [%(default)s]', metavar='INT', default=6)
    parser.add_argument('-n', '--max_length', type=int, help='Maximum length in nucleotides of gene [%(default)s]', metavar='INT', default=10000)
    parser.add_argument('-o', '--outprefix', help='Prefix of output files. If this option is used, a fixed file will be output, together with information on what was changed in the input file. If this option is not used, the script dies if any input sequence is not OK')
    parser.add_argument('infile', help='Input file containing genes to be checked', metavar='Filename')
    options = parser.parse_args()

    pyfastaq.sequences.genetic_code = options.genetic_code
    checker = ariba.refcheck.Checker(
        options.infile,
        min_length=options.min_length,
        max_length=options.max_length
    )
 
    if options.outprefix:
        checker.fix(options.outprefix)
    else:
        ok, reason, seq = checker.check()
        if not ok:
            print('The following sequence not OK, for the reason:', reason)
            print(seq)
            sys.exit(1)



import os
import shutil
import sys
import glob
from setuptools import setup, find_packages


setup(
    name='ariba',
    version='0.4.1',
    description='ARIBA: Antibiotic Resistance Identification By Assembly',
    packages = find_packages(),
    author='Martin Hunt',
    author_email='path-help@sanger.ac.uk',
    url='https://github.com/sanger-pathogens/ariba',
    scripts=glob.glob('scripts/*'),
    test_suite='nose.collector',
    tests_require=['nose >= 1.3'],
    install_requires=[
        'openpyxl',
        'pyfastaq >= 3.0.1',
        'pysam >= 0.8.1',
        'pymummer>=0.0.2'
    ],
    license='GPLv3',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Programming Language :: Python :: 3 :: Only',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    ],
)
